"""Check that JAXGB waveform can be auto-differentiated by JAX."""

import jax
import jax.numpy as jnp
import lisaorbits
from testutils import get_source
from wejax.jacobian import autograd

from jaxgb.jaxgb import JaxGB

# pylint: disable=too-many-locals
jax.config.update("jax_enable_x64", True)


def test_wejax(gw_name: str) -> None:
    """Check that autograd is simply working"""
    size, _, tmax = 2048, 15, 180 * 24 * 3600

    fgb = JaxGB(orbits=lisaorbits.EqualArmlengthOrbits(), t_obs=tmax, t0=0, n=size)

    def tdi(*pars):
        """
        Generate the template frequency serie for the TDI channels
        (A, E).

        Parameters
        ----------
        pars
            template parameters.

        Returns
        -------
           jax.ndarray (Nf x Nc) with the A and E signals (we ignore the T channel)
        """
        A, E, _ = fgb.get_tdi(jnp.array(pars), tdi_combination="AET")
        return jnp.stack([A, E], axis=1)

    params = get_source(gw_name)
    grad_gen = autograd(tdi, argnums=list(range(8)))
    gb_grad = grad_gen(*params)
    assert gb_grad.shape == (size, 2, 8)



"""
JaxGB module
============

Fast waveform and response generator for galactic binaries, as seen by the
LISA instrument. Galactic binaries are described by the following 8
parameters: frequency (f0), frequency evolution (fdot), amplitude
(ampl), sky location (ra, dec), polarisation (psi), inclination
(incl) and initial phase (phi0).

Usage
-----

.. code-block:: python

    # Catalogue parameter definition and definition of epoch `t_init`.
    catalogue_pd = pd.DataFrame(
        {
        "Frequency": #...,
        # other parameters...
        "InitialPhase": #...,
        }
    )
    t_init = 1.0e3  # catalogue epoch

    # Definition of `JaxGB` simulator, with simulation `t0` (simulation epoch).
    t0 = 1.0e4  # simulation t0
    myjaxgb = JaxGB(..., t0=t0, ...)

    # Use GBObject to create the jaxgb input parameter array, which also
    # accounts for the evalution of `f0` and `phi0` for the particular epoch
    # of simulation `t0`.

    # Step 1: define GBObject, holding all source parameters
    catalogue_as_gbo = GBObject.from_pandas_dataframe(catalogue_pd, t_init=t_init)

    # Step 2: get parameters in JaxGB input format, evaluated at `t0`
    params = catalogue_as_gbo.to_jaxgb_array(t0=t0)

    # Step 3: calculate TDI variables in frequency domain
    myjaxgb.get_tdi(jnp.array(params), tdi_generation=1.5, tdi_combination="XYZ")


.. autoclass:: JaxGB
    :members:

.. autoclass:: JaxGBaccurate
    :members:
"""

import jax
import jax.numpy as jnp
import numpy as np
from lisaconstants import c  # pylint: disable=no-name-in-module
from lisaorbits import Orbits
from scipy.signal.windows import tukey

from .tdi import to_tdi_combination, to_tdi_generation


class JaxGB:
    """Galactic binary fast response generation.

    Parameters
    ----------
    orbits
        Orbits instance, based on lisaorbits.
    t_obs
        Duration (s).
    t0
        Start time (s).
    n
        Number of points for slow response evaluation (even-valued integer!).
    """

    def __init__(
        self,
        orbits: Orbits,
        *,
        t_obs: float = 6.2914560e7,
        t0: float = 0.0,
        n: int = 128,
    ) -> None:

        # duration (s)
        self.t_obs = t_obs

        # t0 (s)
        self.t0 = t0

        # number of points for slow response evaluation
        self.n = n

        # time grid
        self.tm = np.linspace(0.0, self.t_obs, num=self.n, endpoint=False)

        # time step slow response
        self.dtm = self.t_obs / self.n

        # initialize orbits and calculate positions of M = 3 spacecraft at time
        # of reception
        self.orbits = orbits
        ltt = self._get_ltt_simple()
        self.arm_length = ltt * c
        self.position = (
            self.orbits.compute_position(self.t0 + self.tm, [1, 2, 3])
            .swapaxes(0, 1)
            .swapaxes(1, 2)
        )  # (3, 3, n)
        self.fstar = c / (self.arm_length * 2 * np.pi)

    def _get_ltt_simple(self) -> jax.Array:
        """Get light-travel times: either pick "L" from orbits file or calculate the an
        ltt average using ``compute_ltt``.

        Returns
        -------
        jax.Array of shape ()
            LTT from orbits file, or averaged calculated LTT (both are scalar
            quantities).
        """

        if hasattr(self.orbits, "L"):
            return jnp.array(self.orbits.L / c)
        if hasattr(self.orbits, "original_attrs"):
            if "L" in self.orbits.original_attrs:
                return jnp.array(self.orbits.original_attrs["L"] / c)
        ltt = self.orbits.compute_ltt(self.t0 + self.tm).T  # (6, n)
        return jnp.mean(ltt)

    def get_kmin(self, f0: jax.Array) -> jax.Array:
        """Calculates kmin, the index of the first heterodyned frequency bin in the
        full frequency grid. Given the observation time baseline `t_obs`, the
        corresponding Fourier frequencies are integer multiples of `1 / t_obs`. The
        nearest frequency bin number to `f0` is therefore `rint(f0 * t_obs)`. This
        bin is centred on the grid of heterodyned Fourier frequencies, of length `n`,
        hence kmin (i.e., the first bin) is shifted by `n/2`.

        Note: `f0` needs to be evaluated at simulation epoch `self.t0`, which in general
        differs from the catalogue epoch `t_init`. Use the :mod:`params.GBObject` to
        calculate source frequencies at simulation epoch `t0` from catalogue frequencies
        (specified at `t_init`).

        Assuming `params` is the `JAXGB`-compatible array of parameters, and `myjaxgb`
        is a `JaxGB` instance, we have:
        .. code-block:: python

            myjaxgb.get_kmin(f0=params[:, 0])
            # `params[:, 0]` needs to contain the catalogue frequencies at the
            # simulation reference time `t0`, which is in general different from the
            # catalogue epoch `t_init`.

        Parameters
        ----------
        f0
            Frequency f0 for source(s), at epoch self.t0 (Hz).

        Returns
        -------
        jax.Array of shape (nsrc,)
            Array with kmin values (integer).
        """

        q = jnp.rint(f0 * self.t_obs)
        return jnp.round(q - self.n / 2).astype(np.int32)

    def get_frequency_grid(self, kmin: jax.Array) -> jax.Array:
        """Return the frequency grid associated to the returned signal.

        Parameters
        ----------
        kmin
            Index of the first frequency bin.

        Returns
        -------
        jax.Array of shape (nsrc, n)
            Array with frequency values (in Hz).
        """
        return (1 / self.t_obs) * (jnp.arange(self.n) + kmin[:, None])

    def _construct_slow_part(
        self,
        f0_fdot_phi0: jax.Array,
        k: jax.Array,
        dpdc: tuple[jax.Array, jax.Array],
        pol_tensors: tuple[jax.Array, jax.Array],
    ) -> jax.Array:
        """This method implements the slow, heterodyned response in time domain. It
        accepts the f0, fdot, and phi0 of all the sources, their direction (k),
        the hplus/hcross strain amplitudes (Dp, Dc) and polarization tensors
        (eplus, ecross). The output is an array of shape (nsrc, 6, n) [other equations
        need to be included here].

        Parameters
        ----------
        f0_fdot_phi0
            array (nsrc, 3) listing f0, fdot, and phi0 values of all sources.
        k
            array (nsrc, 3) with source k-vectors.
        dpdc
            tuple with hplus and hcross strain amplitudes (complex-valued).
        pol_tensors
            tuple with Eplus, Ecross (polarization tensors).

        Returns
        -------
        jax.Array of shape (nsrc, 6, n)
            Time-domain slow heterodyned response.
        """

        nsrc = len(f0_fdot_phi0)

        r = jnp.zeros((4, 3, self.n))  # r convention: 12,13,23,31
        r = r.at[0].set(self.position[1] - self.position[0])  # (3, n)
        r = r.at[1].set(self.position[2] - self.position[0])  # (3, n)
        r = r.at[2].set(self.position[2] - self.position[1])  # (3, n)
        r = r.at[3].set(-r[1])  # (3, n)
        r /= self.arm_length

        # kdotr convention: 12, 21, 13, 31, 23, 32
        kdotr = jnp.zeros((nsrc, 6, self.n))

        kdotr = kdotr.at[:, ::2].set(jnp.dot(k.T, r[:3]))
        kdotr = kdotr.at[:, 1::2].set(-kdotr[:, ::2])

        kdotp = jnp.dot(k.T, self.position)
        kdotp /= c

        f0, fdot, phi0 = f0_fdot_phi0[:, 0], f0_fdot_phi0[:, 1], f0_fdot_phi0[:, 2]
        xi = self.tm - kdotp
        fi = f0[:, None, None] + fdot[:, None, None] * xi
        fonfs = fi / self.fstar  # Ratio of true frequency to transfer frequency

        ### compute transfer f-n
        q = jnp.rint(f0 * self.t_obs)  # index of nearest Fourier bin
        df = 2.0 * np.pi * (q / self.t_obs)
        om = 2.0 * np.pi * f0

        # Reduce to (3, 3, n).
        # A = (eplus @ r * r * dp + ecross @ r * r * dc).sum(axis=2)
        idx = jnp.array([0, 2, 3])
        eplus, ecross = pol_tensors
        aij = (
            jnp.swapaxes(jnp.dot(eplus, r[idx]), 1, 2)
            * r[idx]
            * dpdc[0][:, None, None, None]
            + jnp.swapaxes(jnp.dot(ecross, r[idx]), 1, 2)
            * r[idx]
            * dpdc[1][:, None, None, None]
        )
        asum = aij.sum(axis=2)  # A convention: 12, 23, 31
        arg_s = (phi0[:, None] + (om - df)[:, None] * self.tm[None, :])[:, None, :] + (
            np.pi * fdot[:, None, None] * (xi**2)
        )
        kdotp = om[:, None, None] * kdotp - arg_s

        # gs convention: 12, 23, 31, 21, 32, 13
        idx = jnp.array([0, 4, 3, 1, 5, 2])
        arg_ij = (
            0.5 * fonfs[:, [0, 1, 2, 1, 2, 0]] * (1 + kdotr[:, idx])
        )  # note the plus-sign!
        gs = (
            0.5j
            * (f0[:, None, None] + fdot[:, None, None] * self.tm[None, None, :])
            / self.fstar
            * jnp.sinc(arg_ij / np.pi)
            * jnp.exp(-1j * (arg_ij + kdotp[:, [0, 1, 2, 1, 2, 0]]))
            * asum[:, [0, 1, 2, 0, 1, 2]]
        )  # last term returns a frequency-dependent sign, related to the windowing

        # shift convention: 12, 23, 31, 21, 32, 13 to 12, 23, 31, 13, 32, 21
        gs = gs[:, [0, 1, 2, 5, 4, 3]]

        return gs  # (nsrc, 6, n)

    def _get_gs(self, params: jax.Array) -> jax.Array:
        """Calculate slow response in time domain. Generic for single-link and
        TDI response calculation.

        Parameters
        ----------
        params
            Parameter array of source parameters, as defined in :mod:`jaxgb.params`.

        Returns
        -------
        jax.Array of shape (nsrc, 6, n)
            Slow response in time domain.
        """

        # sign flip phase
        params = params.at[:, 7].set(params[:, 7] * -1)

        # equatorial latitude to theta
        params = params.at[:, 4].set(np.pi / 2 - params[:, 4])

        cosiota = jnp.cos(params[:, 6])
        cosps = jnp.cos(2 * params[:, 5])
        sinps = jnp.sin(2 * params[:, 5])
        aplus = params[:, 2] * (1.0 + cosiota * cosiota)
        across = -2.0 * params[:, 2] * cosiota
        dpdc = (
            aplus * cosps - 1.0j * across * sinps,
            -aplus * sinps - 1.0j * across * cosps,
        )

        sinph = jnp.sin(params[:, 3])
        cosph = jnp.cos(params[:, 3])
        sinth = jnp.sin(params[:, 4])
        costh = jnp.cos(params[:, 4])

        u = jnp.array([costh * cosph, costh * sinph, -sinth], ndmin=2)
        v = jnp.array([sinph, -cosph, jnp.zeros_like(sinph)], ndmin=2)
        k = jnp.array([-sinth * cosph, -sinth * sinph, -costh], ndmin=2)

        eplus = (v[None, :, :] * v[:, None, :] - u[None, :, :] * u[:, None, :]).T
        ecross = (u[None, :, :] * v[:, None, :] + v[None, :, :] * u[:, None, :]).T
        pol_tensors = (eplus, ecross)

        gs = self._construct_slow_part(params[:, [0, 1, 7]], k, dpdc, pol_tensors)

        return gs  # (nsrc, 6, n)

    def _compute_response(self, gs: jax.Array) -> jax.Array:
        """Compute single-link response from gs.

        Parameters
        ----------
        gs
            Slow (heterodyned) response in time domain.

        Returns
        -------
        jax.Array of shape (nsrc, 6, n)
            Frequency-domain response vector.
        """
        yf_slow = jnp.fft.fft(gs, axis=2)

        # yf convention: 12, 23, 31, 13, 32, 21
        yf = jnp.fft.fftshift(yf_slow, axes=2)  # (nsrc, 6, n)
        fctr = 0.5 * self.dtm  # normalization of (half-sided) fourier transform
        return yf * fctr

    def _compute_tdi_xyz(
        self, gs: jax.Array, f0: jax.Array, fdot: jax.Array
    ) -> jax.Array:
        """Compute TDI X, Y, Z from gs.

        Parameters
        ----------
        gs
            Slow (heterodyned) response in time domain.

        Returns
        -------
        jax.Array of shape (nsrc, 3, n)
            Frequency-domain TDI response vector.
        """

        f = f0[:, None] + fdot[:, None] * self.tm[None, :]
        oml = f / np.mean(self.fstar)  # constant arm length assumed
        soml = jnp.sin(oml)
        fctr = jnp.exp(-1.0j * oml)
        fctr2 = -2.0j * soml * fctr

        # gs convention: 12, 23, 31, 13, 32, 21
        xsl = gs[:, 5] - gs[:, 2] + (gs[:, 0] - gs[:, 3]) * fctr
        ysl = gs[:, 4] - gs[:, 0] + (gs[:, 1] - gs[:, 5]) * fctr
        zsl = gs[:, 3] - gs[:, 1] + (gs[:, 2] - gs[:, 4]) * fctr

        xyzsl = fctr2[:, None, :] * jnp.concatenate(
            [xsl[:, None, :], ysl[:, None, :], zsl[:, None, :]], axis=1
        )
        xyzf_slow = jnp.fft.fft(xyzsl, axis=2)
        xyzf = jnp.fft.fftshift(xyzf_slow, axes=2)

        fctr3 = 0.5 * self.dtm
        xyzf *= fctr3
        return xyzf

    def get_link_responses(self, params: jax.Array) -> jax.Array:
        """Calculate the 6 single-link responses in frequency domain.

        Parameters
        ----------
        params
            Parameter array of shape (nsrc, 8) of source parameters, as
            defined in :mod:`jaxgb.params`.

        Returns
        -------
        jax.Array of shape (nsrc, 6, n)
            Single-link response vector.
        """
        params = params.copy()
        gs = self._get_gs(params)
        yf = self._compute_response(gs)
        return yf

    def get_tdi(
        self,
        params: jax.Array,
        tdi_generation: float = 1.5,
        tdi_combination: str = "XYZ",
    ) -> tuple[jax.Array, jax.Array, jax.Array]:
        """Calculate TDI XYZ in frequency domain.

        Parameters
        ----------
        params
            Parameter array of shape (nsrc, 8) of source parameters, as
            defined in :mod:`jaxgb.params`.
        tdi_generation
            TDI generation (either 1.5 or 2).
        tdi_combination
            TDI combinations (either "XYZ" or "AET").

        Returns
        -------
        tuple of arrays, each of shape (nsrc, n)
            TDI frequency-domain response channels (X, Y, Z) or
            (A, E, T).
        """
        _params = params.copy()
        if len(_params.shape) == 1:
            _params = _params.reshape(1, -1)
        gs = self._get_gs(_params)
        tdi = self._compute_tdi_xyz(gs, _params[:, 0], _params[:, 1])

        kmin = self.get_kmin(f0=_params[:, 0])
        freqs = self.get_frequency_grid(kmin)

        tdi = to_tdi_generation(
            tdi, freqs, to=tdi_generation, arm_length=self.arm_length
        )
        tdi = to_tdi_combination(tdi, tdi_combination)
        x = tdi[:, 0, :]
        y = tdi[:, 1, :]
        z = tdi[:, 2, :]
        if len(params.shape) == 1:
            return x[0], y[0], z[0]
        return x, y, z

    def sum_tdi(
        self,
        params: jax.Array,
        kmin: int | None = None,
        kmax: int | None = None,
        **kwargs,
    ) -> tuple[jax.Array, jax.Array, jax.Array]:
        r"""Sum TDI response channels for a set of source parameters.

        Each source is simulated independently, and the TDI responses
        are then summed (after aligning the frequency grids) and
        returned.

        Parameters
        ----------
        params
            Parameter array with shape (nsrc, 8) of source parameters,
            as defined in :mod:`jaxgb.params`.
        kmin
            First index of the frequency grid. If None, takes the
            first index of the lowest frequency source.
        kmax
            Last index of the frequency grid. If None, takes the
            last index of the highest frequency source.
        kwargs
             Extra keywords arguments passed to :meth:`get_tdi`.

        Returns
        -------
        tuple (jax.Array, jax.Array, jax.Array, ), each of shape (nfreq)
            TDI frequency-domain response channels (X, Y, Z) or
            (A, E, T) [strain].

        """
        kmins = self.get_kmin(f0=params[:, 0])
        kmin = jnp.min(kmins).item() if kmin is None else kmin
        kmax = jnp.max(kmins).item() + self.n if kmax is None else kmax

        def _add_to(tdi_one, tdi_sum, _kmin):
            def _update_fpoint(idx, val):
                _mask = ((idx + _kmin) >= 0) & ((idx + _kmin) < kmax)
                return jax.lax.dynamic_update_index_in_dim(
                    val,
                    val[:, idx + _kmin] + _mask * tdi_sum[:, idx],
                    idx + _kmin,
                    axis=1,
                )

            return jax.lax.fori_loop(0, self.n, _update_fpoint, tdi_one)

        def _add_one(idx, tdi):
            _x, _y, _z = self.get_tdi(
                jax.lax.dynamic_slice_in_dim(params, idx, 1, axis=0),
                **kwargs,
            )
            return _add_to(tdi, jnp.vstack([_x, _y, _z]), kmins[idx] - kmin)

        _tdi = jnp.zeros((3, kmax - kmin), dtype=np.complex128)
        _tdi = jax.lax.fori_loop(0, params.shape[0], _add_one, _tdi)
        return tuple(_tdi)


class JaxGBaccurate(JaxGB):
    """Galactic binary less fast but more accurate response generation. Considers
    time-varying arm lengths and reduces edge artefacts (ringing).

    Parameters
    ----------

    window
        Dimensionless extension factor for slow response calculation, for
        windowing purposes. The slow response will be calculated on the domain
        [t0 - window * t_obs, t0 + t_obs + window * t_obs], i.e., the extension is
        symmetric around the time interval of interest [t0, t0 + t_obs].

        window needs to have a strictly positive integer or half-integer value. Note
        that using window also modifies the kmin values:
        df = 1/(t_obs + 2 * window * t_obs).
    """

    def __init__(self, window: float = 0.5, **kwargs) -> None:
        super().__init__(**kwargs)

        # window value (strictly positive integer or half-integer)
        self.window = window

        # length of extended time grid (with windowing)
        self.t_ext = self.t_obs + 2.0 * self.window * self.t_obs

        # time grid with windowing [-window*t_obs, t_obs + 2*window*t_obs]
        self.tm = (
            np.linspace(
                0,
                self.t_ext,
                num=self.n,
                endpoint=False,
            )
            - self.window * self.t_obs
        )

        # time step slow response
        self.dtm = self.t_ext / self.n

        # calculate positions of M = 3 spacecraft at time of emission
        ltt = self.orbits.compute_ltt(self.t0 + self.tm).T  # (6, n), recalculate
        self.position = (
            self.orbits.compute_position(self.t0 + self.tm, [1, 2, 3])
            .swapaxes(0, 1)
            .swapaxes(1, 2)
        )  # (M, 3, n)
        self.positionemi = jnp.array(
            [
                self.orbits.compute_position(self.t0 + self.tm - ltt[0], [2])
                .squeeze()
                .T,
                self.orbits.compute_position(self.t0 + self.tm - ltt[1], [3])
                .squeeze()
                .T,
                self.orbits.compute_position(self.t0 + self.tm - ltt[2], [1])
                .squeeze()
                .T,
                self.orbits.compute_position(self.t0 + self.tm - ltt[3], [3])
                .squeeze()
                .T,
                self.orbits.compute_position(self.t0 + self.tm - ltt[4], [2])
                .squeeze()
                .T,
                self.orbits.compute_position(self.t0 + self.tm - ltt[5], [1])
                .squeeze()
                .T,
            ]
        )  # (6, 3, n)
        self.arm_length = ltt * c
        fstar = c / (self.arm_length * 2 * np.pi)
        self.fstar = fstar[None, :, :]
        # convention: 12, 23, 31, 13, 32, 21

    def get_kmin(self, f0: jax.Array) -> jax.Array:
        """Calculates kmin, the index of the first heterodyned frequency bin in the
        full frequency grid. For windowed response calculations, the observation time
        baseline is extended from `t_obs` to `t_ext = t_obs + 2 * window * t_obs`, so
        the corresponding Fourier frequencies are integer multiples of `1 / t_ext`.
        The nearest frequency bin number to `f0` is therefore `rint(f0 * t_ext)`. This
        bin is centred on the grid of heterodyned Fourier frequencies, of length `n`,
        hence kmin (i.e., the first bin) is shifted by `n/2`.

        If `t_init` is different from `t0`, `f0` needs to be corrected for the linear
        drift proportional to `fdot`.

        Parameters
        ----------
        f0
            Frequency f0 for source(s) (Hz)

        Returns
        -------
        jax.Array of shape (nsrc,)
            Array with kmin values (integer)
        """
        q = jnp.rint(f0 * self.t_ext)
        return jnp.round(q - self.n / 2).astype(np.int32)

    def get_frequency_grid(self, kmin: jax.Array) -> jax.Array:
        """Return the frequency grid associated to the returned signal.

        Parameters
        ----------
        kmin
            Index of the first frequency bin.

        Returns
        -------
        jax.Array of shape (nsrc, n)
            Array with frequency values (in Hz).
        """
        return (1 / self.t_ext) * (jnp.arange(self.n) + kmin[:, None])

    def _construct_slow_part(
        self,
        f0_fdot_phi0: jax.Array,
        k: jax.Array,
        dpdc: tuple[jax.Array, jax.Array],
        pol_tensors: tuple[jax.Array, jax.Array],
    ) -> jax.Array:
        """This method implements the slow, heterodyned response in time domain. It
        accepts the f0, fdot, and phi0 of all the sources, their direction (k),
        the hplus/hcross strain amplitudes (Dp, Dc) and polarization tensors
        (eplus, ecross). The output is an array of shape  (nsrc, 6, n) [other equations
        need to be included here].

        For this implementation, we include tukey windowing immediately here, which
        tapers off the slow response at the edges of the extended domain.

        Parameters
        ----------
        f0_fdot_phi0
            array (nrsc, 3) listing f0, fdot, and phi0 values of all sources.
        k
            array (nsrc, 3) with source k-vectors.
        dpdc
            tuple with hplus and hcross strain amplitudes (complex-valued).
        pol_tensors
            tuple with Eplus, Ecross (polarization tensors).

        Returns
        -------
        jax.Array of shape (nsrc, 6, n)
            Time-domain slow heterodyned response.
        """

        r = jnp.zeros((6, 3, self.n))  # r convention: 12, 23, 31, 13, 32, 21
        r = r.at[0].set(self.position[0] - self.positionemi[0])  ## [3xn]
        r = r.at[1].set(self.position[1] - self.positionemi[1])  ## [3xn]
        r = r.at[2].set(self.position[2] - self.positionemi[2])  ## [3xn]
        r = r.at[3].set(self.position[0] - self.positionemi[3])  ## [3xn]
        r = r.at[4].set(self.position[2] - self.positionemi[4])  ## [3xn]
        r = r.at[5].set(self.position[1] - self.positionemi[5])  ## [3xn]
        r /= self.arm_length[:, None, :]

        # kdotr convention: 12, 23, 31, 13, 32, 21
        kdotr = jnp.dot(k.T, r)
        kdotp = jnp.dot(k.T, self.position)
        kdotp /= c

        f0, fdot, phi0 = f0_fdot_phi0[:, 0], f0_fdot_phi0[:, 1], f0_fdot_phi0[:, 2]
        xi = self.tm - kdotp
        fi = f0[:, None, None] + fdot[:, None, None] * xi
        # Ratio of true frequency to transfer frequency
        fonfs = fi[:, [0, 1, 2, 0, 2, 1]] / self.fstar

        ### compute transfer f-n
        q = jnp.rint(f0 * self.t_ext)  # index of nearest Fourier bin
        df = 2.0 * np.pi * (q / self.t_ext)
        om = 2.0 * np.pi * f0

        # Reduce to (3, 3, n).
        # A = (eplus @ r * r * dp + ecross @ r * r * dc).sum(axis=2)
        eplus, ecross = pol_tensors
        aij = (
            jnp.swapaxes(jnp.dot(eplus, r), 1, 2) * r * dpdc[0][:, None, None, None]
            + jnp.swapaxes(jnp.dot(ecross, r), 1, 2) * r * dpdc[1][:, None, None, None]
        )
        asum = aij.sum(axis=2)  # A convention: 12, 23, 31, 13, 32, 21
        arg_s = (phi0[:, None] + (om - df)[:, None] * self.tm[None, :])[:, None, :] + (
            np.pi * fdot[:, None, None] * (xi**2)
        )
        kdotp = om[:, None, None] * kdotp - arg_s

        # gs convention: 12, 23, 31, 13, 32, 21
        idx = jnp.array([0, 1, 2, 0, 2, 1])
        arg_ij = 0.5 * fonfs * (1 - kdotr)  # note the minus-sign!
        gs = (
            0.5j
            * (f0[:, None, None] + fdot[:, None, None] * self.tm[None, None, :])
            / self.fstar
            * jnp.sinc(arg_ij / np.pi)
            * jnp.exp(-1j * (arg_ij + kdotp[:, idx]))
            * asum
            * jnp.exp(
                -2.0j * np.pi * q[:, None, None] * self.window * self.t_obs / self.t_ext
            )
        )  # last term returns a frequency-dependent sign, related to the windowing

        # finally, apply windowing
        w = tukey(len(self.tm), alpha=0.45) ** 4
        gs *= w[None, None, :]

        return gs  # (nsrc, 6, n)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# pylint: disable=all

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys

sys.path.insert(0, os.path.abspath("."))
sys.path.append(os.path.abspath("./_ext"))
import jaxgb

# -- Project information -----------------------------------------------------

project = "JAXGB"
version = jaxgb.__version__
release = jaxgb.__version__
copyright = jaxgb.__copyright__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx_rtd_theme",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.autosummary",
]

autosummary_generate = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

intersphinc_mapping = {
    "python": ("https://docs.python.org/3", None),
    "jax": ("https://jax.readthedocs.io/en/latest", None),
}


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]

# html_logo = "_static/img/logo.svg"
# html_favicon = "_static/img/favicon.ico"
html_show_sphinx = False

# These paths are either relative to html_static_path
# or fully qualified paths (eg. https://...)
# html_css_files = [ ]

# -- Options for extensions --------------------------------------------------

typehints_defaults = "comma"
always_use_bars_union = True

autodoc_type_aliases = {
    "Iterable": "Iterable",
    "ArrayLike": "ArrayLike",
}

autosectionlabel_prefix_document = True

mathjax3_config = {
    "loader": {"load": ["[tex]/physics"]},
    "tex": {
        "packages": {"[+]": ["physics"]},
    },
}

bibtex_bibfiles = ["bibliography.bib"]
bibtex_default_style = "unsrt"



"""
Test TDI computation of several sources at one time, either with
jax.vmap or multi columns parameters
"""

import jax
import jax.numpy as jnp
import lisaorbits
import numpy as np
import pandas as pd

from jaxgb.jaxgb import JaxGB
from jaxgb.params import GBObject

jax.config.update("jax_enable_x64", True)


def test_multicol(catalog: pd.DataFrame, tdi_generation: float) -> None:
    """Test the computation of tdi for several sources at one time, using
    2D input array.
    """

    size, tmax = 512, 180 * 24 * 3600
    ncol = len(catalog)
    fgb = JaxGB(orbits=lisaorbits.EqualArmlengthOrbits(), t_obs=tmax, t0=0, n=size)

    # create JAXGB compatible input array from catalog
    params = GBObject.from_pandas_dataframe(catalog, t_init=0.0).to_jaxgb_array(t0=0.0)

    X, _, _ = fgb.get_tdi(params, tdi_generation=tdi_generation)
    assert X.shape == (ncol, size)

    i_check = 10
    x, _, _ = fgb.get_tdi(params[i_check], tdi_generation=tdi_generation)
    assert np.isclose(x, X[i_check], atol=1e-25).all()


def test_vmap(catalog: pd.DataFrame) -> None:
    """Test the computation of tdi for several sources at one time, using
    jax.vmap
    """

    size, tmax = 512, 180 * 24 * 3600
    ncol = len(catalog)
    fgb = JaxGB(orbits=lisaorbits.EqualArmlengthOrbits(), t_obs=tmax, t0=0, n=size)

    # create JAXGB compatible input array from catalog
    params = GBObject.from_pandas_dataframe(catalog, t_init=0.0).to_jaxgb_array(t0=0.0)

    X1, _, _ = jax.vmap(fgb.get_tdi, in_axes=0)(params)
    assert X1.shape == (ncol, size)

    X2, _, _ = fgb.get_tdi(params)
    assert np.isclose(X1, X2, atol=1e-25).all()


def test_sum(catalog: pd.DataFrame) -> None:
    """Test the computation of the summed tdi for several sources at one time"""

    size, tmax = 512, 180 * 24 * 3600

    # create JAXGB compatible input array from catalog
    params = GBObject.from_pandas_dataframe(catalog, t_init=0.0).to_jaxgb_array(t0=0.0)

    fgb = JaxGB(orbits=lisaorbits.EqualArmlengthOrbits(), t_obs=tmax, t0=0, n=size)

    # test for one source
    i_check = 10
    x, _, _ = fgb.get_tdi(params[i_check])
    X, _, _ = fgb.sum_tdi(jnp.vstack([params[i_check], params[i_check]]))
    assert np.isclose(X, 2 * x, atol=1e-25).all()

    X1, _, _ = fgb.sum_tdi(params, kmin=0, kmax=100_000)
    assert X1.shape == (100_000,)

    # assuming no overlap of the highest freq source with any other
    catalog = catalog.sort_values(by=["Frequency"])
    params = GBObject.from_pandas_dataframe(catalog, t_init=0.0).to_jaxgb_array(t0=0.0)
    last = params[len(catalog) - 1, 0]
    x, _, _ = fgb.get_tdi(params[len(catalog) - 1])  # select last source
    kmin = fgb.get_kmin(last)  # params[i_last, 0] = f0
    X, _, _ = fgb.sum_tdi(params, kmin=kmin, kmax=kmin + size)
    assert fgb.get_kmin(params[len(catalog) - 2, 0]) + size < kmin
    assert np.isclose(X, x, atol=1e-25).all()



"""JAXGB"""

from importlib.metadata import PackageNotFoundError, metadata, version

try:
    __version__ = version(__name__)
    __license__ = metadata(__name__)["license"]
except PackageNotFoundError:
    __version__ = "0.0.0"  # Fallback for development mode
    __license__ = "Undefined"

__copyright__ = "2023, CEA and CNRS"



"""
pytest fixtures definition.
"""

import lisaorbits
import numpy as np
import pandas as pd
import pytest


@pytest.fixture(params=[1.5, 2])
def tdi_generation(request):
    """TDI generation options.
    For now a simple scalar.
    """
    return request.param


@pytest.fixture(params=["equal", "keplerian"])
def orbits_approximation(tmp_path, request):
    """Analytic orbits provider, build and save on disk."""
    tmax = 730 * 24 * 3600
    if request.param == "equal":
        orbits = lisaorbits.EqualArmlengthOrbits()
    else:
        orbits = lisaorbits.KeplerianOrbits()
    orbits_fn = str(tmp_path / "orbits.h5")
    orbits.write(orbits_fn, mode="w", dt=1e4, size=(3 * tmax) // 1e4, t0=-2 * tmax)
    return orbits_fn


@pytest.fixture
def esa_orbits(tmp_path):
    """ESA orbits provider, build and save on disk."""
    tmax = 730 * 24 * 3600
    orbits = lisaorbits.OEMOrbits.from_included("esa-leading")
    orbits.t_init = orbits.t_init + 50000000
    orbits_fn = str(tmp_path / "orbits.h5")
    orbits.write(
        orbits_fn, mode="w", dt=1e5, size=(2 * tmax) // 1e5, t0=orbits.t_init - 100
    )
    return orbits_fn


@pytest.fixture
def gw_name():
    """List of GW short name. See testutils.get_source."""
    return "highfreq"


@pytest.fixture
def catalog():
    """Return a random catalog of 100 sources in pandas.DataFrame format."""
    ncol = 100  # number of sources
    freqs = np.sort(1e-3 * np.abs(np.random.randn(ncol)))
    freqs[-1] = freqs[-2] + 1e-3  # 1mHz offset
    freqs = freqs[np.random.choice(np.arange(ncol), size=ncol, replace=False)]
    _catalog = pd.DataFrame(
        {
            "Frequency": freqs,
            "FrequencyDerivative": 1e-13 * np.abs(np.random.randn(ncol)),
            "Amplitude": 1e-22 * np.abs(np.random.randn(ncol)),
            "Inclination": np.arccos(np.random.uniform(size=ncol, low=-1.0, high=1.0)),
            "RightAscension": np.random.uniform(size=ncol, low=0, high=2 * np.pi),
            "Declination": np.random.uniform(size=ncol, low=-np.pi / 2, high=np.pi / 2),
            "Polarization": np.random.uniform(size=ncol, low=0, high=2 * np.pi),
            "InitialPhase": np.random.uniform(size=ncol, low=0, high=2 * np.pi),
        }
    )
    return _catalog



"""
JaxGB TDI module
================

This is a placeholder waiting for a third-party software managing TDI
combinations and generations.

For now it gives basic transformations from XYZ (Michelson) to AET
(orthogonal) assuming equal diagonal and cross terms in the noise matrix,
and basic transformation from TDI 1.5 to TDI 2 assuming equal
armlengths.

.. autofunction:: to_tdi_generation
.. autofunction:: to_tdi_combination
"""

import jax
import jax.numpy as jnp
import numpy as np
from lisaconstants import c  # pylint: disable=no-name-in-module


def to_tdi_generation(
    tdi: jax.Array,
    f: jax.Array,
    to: float = 1.5,
    arm_length: jax.Array = jnp.array(2.5e9),
) -> jax.Array:
    """Frequency-domain transformation of TDI 1.5 to another generation.

    The transformation is done in the frequency domain, assuming equal
    armlengths, ie., a single factor.

    Parameters
    ----------
    tdi
        Input TDI in the frequency domain of shape
        (nsource, 3, nfreq).
    f
        Frequency of shape nfreq [Hz]
    to
        either 1.5 or 2
    arm_length
        Constant armlengths [m].

    Returns
    -------
    Array of shape (nsource, 3, freq)
        Output TDI in the frequency domain.
    """
    if to == 2:
        omegal = 2 * np.pi * f * jnp.mean(arm_length) / c
        fctr = 2.0j * jnp.sin(2 * omegal) * jnp.exp(-2j * omegal)
        return tdi * fctr[:, None, :]
    return tdi


def to_tdi_combination(tdi: jax.Array, combination: str = "XYZ") -> jax.Array:
    """Transform input TDI XYZ combination into given combination.

    This can be applied to time or frequency series.

    Parameters
    ----------
    tdi
        Input TDI of shape (nsource, 3, nfreq).
    combination
        Combination name, can be XYZ or AET.

    Returns
    -------
    Array of shape (nsource, 3, freq)
        Output TDI.
    """
    if combination == "AET":
        X, Y, Z = tdi[:, 0, :], tdi[:, 1, :], tdi[:, 2, :]
        A, E, T = (
            (Z - X) / np.sqrt(2.0),
            (X - 2.0 * Y + Z) / np.sqrt(6.0),
            (X + Y + Z) / np.sqrt(3.0),
        )
        tdi = jnp.stack([A, E, T], axis=1)
        return tdi
    return tdi



#
# BSD 3-Clause License
#
# Copyright (c) 2022, California Institute of Technology and
# Max Planck Institute for Gravitational Physics (Albert Einstein Institute)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# This software may be subject to U.S. export control laws. By accepting this
# software, the user agrees to comply with all applicable U.S. export laws and
# regulations. User has the responsibility to obtain export licenses, or other
# export authority as may be required before exporting such information to
# foreign countries or providing access to foreign persons.
#
"""
Script to upload and publish a new version on Zenodo.

Prior to running this script run `python -m build` to build a distribution archive for
the package that is meant to be uploaded to Zenodo.
"""

import argparse
import json
import logging
import os
from datetime import datetime

import requests

logger = logging.getLogger(__name__)


def upload_and_publish(recid, token, version, publish, path, sandbox):
    """Upload and publish a new version on Zenodo.

    Args:
        recid (int): Zenodo ID of the first version
        token (str): Zenodo access token
        version (str): new deposition version
        publish (bool): whether new record should be published
        path (str): path to archive to be uploaded
        sandbox (bool): use Zenodo sandbox
    """
    # pylint: disable=too-many-statements
    params = {"access_token": token}
    headers = {"Content-Type": "application/json"}

    publication_date = datetime.today().strftime("%Y-%m-%d")

    url = "https://sandbox.zenodo.org" if sandbox else "https://zenodo.org"

    # Retrieve latest deposition and read out its id
    try:
        req = requests.get(
            f"{url}/api/deposit/depositions/{recid}", params=params, timeout=20
        )
        req.raise_for_status()
    except requests.exceptions.HTTPError as error:
        raise RuntimeError("Retrieval of latest deposition failed") from error
    logger.info("Request to deposition '%d' successful", recid)

    try:
        req = requests.get(req.json()["links"]["latest"], timeout=20)
        req.raise_for_status()
    except requests.exceptions.HTTPError as error:
        raise RuntimeError("Retrieval of latest deposition ID failed") from error
    latest_id = req.json()["id"]
    logger.info("Obtained latest deposition ID '%d'", latest_id)

    # Create new version
    try:
        req = requests.post(
            f"{url}/api/deposit/depositions/{latest_id}/actions/newversion",
            params=params,
            timeout=20,
        )
        req.raise_for_status()
    except requests.exceptions.HTTPError as error:
        raise RuntimeError("Creation of new version failed") from error
    draft_url = req.json()["links"]["latest_draft"]
    logger.info("Created new draft version at '%s'", draft_url)

    def _write_metadata(path):
        """Write metadata of the draft deposition."""
        try:
            with open(path, encoding="utf-8") as file:
                metadata = json.load(file)

            metadata["version"] = version
            metadata["publication_date"] = publication_date
            data = {"metadata": metadata}

            req = requests.put(
                draft_url,
                params=params,
                data=json.dumps(data),
                headers=headers,
                timeout=20,
            )
            req.raise_for_status()
        except (FileNotFoundError, requests.exceptions.HTTPError) as error:
            raise RuntimeError(
                "Writing of metadata failed, "
                "consider removing new draft version manually online"
            ) from error

    # Write metadata
    if sandbox:
        # reduced set of metadata since communities are not available on Zenodo sandbox
        _write_metadata(".zenodo_sandbox.json")
    else:
        _write_metadata(".zenodo.json")
    logger.info("Wrote metadata to draft")

    # Delete all files
    try:
        req = requests.get(draft_url + "/files", params=params, timeout=20)
        for file in req.json():
            file_url = file["links"]["self"]
            req = requests.delete(file_url, params=params, timeout=20)
        req.raise_for_status()
    except requests.exceptions.HTTPError as error:
        raise RuntimeError(
            "Deletion of files failed, consider "
            "removing new draft version manually online"
        ) from error
    logger.info("Deleted old files from draft")

    # Upload new files
    try:
        with open(path, "rb") as file:
            data = {"name": os.path.basename(path)}
            files = {"file": file}
            req = requests.post(
                draft_url + "/files", params=params, data=data, files=files, timeout=20
            )
            req.raise_for_status()
    except (FileNotFoundError, requests.exceptions.HTTPError) as error:
        raise RuntimeError("File upload failed") from error
    logger.info("Upload files to draft")

    # Publish new version
    if publish:
        try:
            req = requests.post(
                draft_url + "/actions/publish", params=params, timeout=20
            )
        except requests.exceptions.HTTPError as error:
            raise RuntimeError(
                "Publication failed, consider removing new draft "
                "version manually online"
            ) from error
        logger.info("Published draft")


def main():
    """Parse command-line arguments and upload and publish."""
    parser = argparse.ArgumentParser(description="Upload and publish to Zenodo.")
    parser.add_argument("--project", type=int, help="Zenodo ID")
    parser.add_argument("--token", help="Zenodo access token")
    parser.add_argument("--version", help="version of the new deposition")
    parser.add_argument(
        "--publish", action="store_true", help="whether new record should be published"
    )
    parser.add_argument(
        "--sandbox", action="store_true", help="use Zenodo sandbox for testing"
    )
    parser.add_argument("path", help="path to archive")

    args = parser.parse_args()

    upload_and_publish(
        args.project, args.token, args.version, args.publish, args.path, args.sandbox
    )


if __name__ == "__main__":
    main()



"""
JaxGB parameter module
======================

This is a module clarifying the parameters (names, units, and LaTeX notation) and
their ordering in JaxGB. It also enables conversions to/from the parameterisation
used in other tools, by defining a `GBObject` containing all necessary parameters.

.. autoclass:: GBObject
    :members:
"""

import jax
import jax.numpy as jnp
import numpy as np
import pandas as pd
from astropy.coordinates import SkyCoord

PARAM_NAMES = np.array(
    [
        "Frequency",
        "Frequency derivative",
        "Amplitude",
        "Right ascension",
        "Declination",
        "Polarisation angle",
        "Inclination",
        "Initial phase",
    ]
)
"""Array of ordered parameter names."""

PARAM_LATEX = np.array(
    [
        "$f_0$",
        r"$\dot{f}$",
        "$A$",
        r"$\mathrm{RA}$",
        r"$\mathrm{dec}$",
        r"$\psi$",
        r"$\iota$",
        r"$\phi_0$",
    ]
)
"""Array of ordered Latex-formatted parameter symbols."""

PARAM_UNITS = np.array(
    [
        r"$\mathrm{Hz}$",
        r"$\mathrm{Hz}/\mathrm{s}$",
        "",
        r"$\mathrm{rad}$",
        r"$\mathrm{rad}$",
        r"$\mathrm{rad}$",
        r"$\mathrm{rad}$",
        r"$\mathrm{rad}$",
    ]
)
"""Array of ordered parameter expected units."""


class GBObject:
    """Circular Galactic binary parameter set.

    GBObject brings together a canonical parameterisation of a set of circular
    Galactic Binaries. It enables conversion into the parameterisation used by
    different (circular) Galactic Binary response codes, like LISA GW Response, LDC
    tools, fastgb, and gbgpu. At the moment, we ignore the inclusion of a second
    derivative of the frequency.

    Parameters
    ----------
    f0
        Gravitational-wave signal frequency at epoch [Hz].
    fdot
        Gravitational-wave signal frequency derivative [Hz/s].
    A
        Gravitational-wave signal amplitude [strain].
    ra
        Right ascension (equatorial longitude) [rad].
    dec
        Declination, equatorial latitude [rad].
    psi
        Polarisation angle [rad].
    iota
        Inclination angle [rad].
    phi0
        Initial phase at epoch [rad].
    t_init
        Epoch used to define the initial phase phi0 and the frequency f0, in TCB [s].

    """

    def __init__(
        self,
        *,
        f0: np.ndarray,
        fdot: np.ndarray,
        A: np.ndarray,
        ra: np.ndarray,
        dec: np.ndarray,
        psi: np.ndarray,
        iota: np.ndarray,
        phi0: np.ndarray,
        t_init: float,
    ) -> None:

        self.f0 = f0
        """Gravitational-wave signal frequency at epoch [Hz]."""
        self.fdot = fdot
        """Gravitational-wave signal frequency derivative [Hz/s]."""
        self.A = A
        """Gravitational-wave signal amplitude [strain]."""
        self.ra = ra
        """Right ascension (equatorial longitude) [rad]."""
        self.dec = dec
        """Declination, equatorial latitude [rad]."""
        self.psi = psi
        """Polarisation angle [rad]."""
        self.iota = iota
        """Inclination angle [rad]."""
        self.phi0 = phi0
        """Initial phase at epoch [rad]."""
        self.t_init = t_init
        """Epoch used to define the initial phase phi0 and the frequency f0, in TCB
        [s]."""

    @classmethod
    def from_jaxgb_params(cls, jaxgb_params: np.ndarray, t_init: float) -> "GBObject":
        """Instantiate GB from a JaxGB params array.

        Parameters
        ----------
        jaxgb_params : array-like of shape (nsrc, 8)
            Parameters, as defined in :attr:`PARAM_NAMES` and units
            as defined in :attr:`PARAM_UNITS`.
        t_init : float
            Epoch used to define the initial phase phi0 and the frequency f0, in TCB
            [s].

        Returns
        -------
            A new GB instance.
        """
        gb_object = cls(
            f0=jaxgb_params[:, 0],
            fdot=jaxgb_params[:, 1],
            A=jaxgb_params[:, 2],
            ra=jaxgb_params[:, 3],
            dec=jaxgb_params[:, 4],
            psi=jaxgb_params[:, 5],
            iota=jaxgb_params[:, 6],
            phi0=jaxgb_params[:, 7],
            t_init=t_init,
        )
        return gb_object

    @classmethod
    def from_pandas_dataframe(
        cls, pandas_df: pd.DataFrame, t_init: float
    ) -> "GBObject":
        """Instantiate GB from a Pandas dataframe.

        Parameters
        ----------
        pandas_df
            Pandas dataframe of parameters.
        t_init : float
            Epoch used to define the initial phase phi0 and the frequency f0, in TCB
            [s].

        Returns
        -------
            A new GB instance.
        """
        gb_object = cls(
            f0=pandas_df["Frequency"].to_numpy(),
            fdot=pandas_df["FrequencyDerivative"].to_numpy(),
            A=pandas_df["Amplitude"].to_numpy(),
            ra=pandas_df["RightAscension"].to_numpy(),
            dec=pandas_df["Declination"].to_numpy(),
            psi=pandas_df["Polarization"].to_numpy(),
            iota=pandas_df["Inclination"].to_numpy(),
            phi0=pandas_df["InitialPhase"].to_numpy(),
            t_init=t_init,
        )
        return gb_object

    @classmethod
    def from_records(cls, rec_array: np.recarray | dict, t_init: float) -> "GBObject":
        """Instantiate GB from a numpy record array.

        Parameters
        ----------
        rec_array
            Numpy record array or dictionnary of parameters.
        t_init : float
            Epoch used to define the initial phase phi0 and the frequency f0, in TCB
            [s].

        Returns
        -------
            A new GB instance.
        """
        gb_object = cls(
            f0=rec_array["Frequency"],
            fdot=rec_array["FrequencyDerivative"],
            A=rec_array["Amplitude"],
            ra=rec_array["RightAscension"],
            dec=rec_array["Declination"],
            psi=rec_array["Polarization"],
            iota=rec_array["Inclination"],
            phi0=rec_array["InitialPhase"],
            t_init=t_init,
        )
        return gb_object

    @staticmethod
    def _equatorial_to_ecliptic(ra: float, dec: float) -> tuple[float, float]:
        """Convert (ra, dec) into ecliptic longitude and lattitude (lambda, beta).

        Parameters
        ----------
        ra
            Right ascension [rad].
        dec
            Declination [rad].

        Returns
        -------
        tuple (lambda, beta)
            Ecliptic longitude and latitude [rad, rad].
        """
        sky_icrs = SkyCoord(ra=ra, dec=dec, frame="icrs", unit="rad")
        ecl_coords = sky_icrs.barycentrictrueecliptic
        return ecl_coords.lon.rad, ecl_coords.lat.rad

    def _to_array_at_reference_time(self, t0: float) -> np.ndarray:
        """Create a numpy array with parameter at reference time `t0`.

        Returns
        -------
        np.array of shape (nsrc, 8)
            Array of parameters in the same order as :attr:`PARAM_NAMES`
            and with units :attr:`PARAM_UNITS`.
        """
        # shift frequency `f0`, which evolves at the linear rate `fdot`
        f0_t0 = self.f0 - self.fdot * (self.t_init - t0)

        # shift `phi0`, to account for difference `t_init` (catalogue epoch) and `t0`
        _shift = np.pi * (
            2 * self.f0 * (self.t_init - t0) - self.fdot * (self.t_init - t0) ** 2
        )
        phi0_t0 = (self.phi0 + _shift) % (2 * np.pi)

        return np.array(
            [
                f0_t0,  # shifted f0
                self.fdot,
                self.A,
                self.ra,
                self.dec,
                self.psi,
                self.iota,
                phi0_t0,  # shifted phi0
            ]
        ).T

    def to_jaxgb_array(self, t0: float) -> jax.Array:
        """Create JaxGB array, respecting format and ordering.

        Parameters
        ----------
        t0
            Starting time of `JAXGB` response calculation in TCB (s). This time is used
            to shift the GW frequency `f0` and the phase `phi0`, defined at `t_init`, to
            the simulation starting time.

        Returns
        -------
        jax.Array of shape (nsrc, 8)
            Array of parameters in the same order as :attr:`PARAM_NAMES`
            and with units :attr:`PARAM_UNITS`.
        """
        return jnp.array(self._to_array_at_reference_time(t0=t0))

    def to_ldc_dict(self) -> dict[str, np.ndarray]:
        """Create LDC dictionary, respecting format and ordering.

        Sky positions are converted to ecliptic coordinates. Parameters are specified
        at the epoch `t_init`.

        Returns
        -------
            Dictionary of parameters.
        """
        # LDC assumes no difference between `t0` and `t_init`
        params = self._to_array_at_reference_time(t0=self.t_init)

        # convert equatorial to ecliptic coordinates, change order
        params[4], params[3] = GBObject._equatorial_to_ecliptic(params[3], params[4])

        return dict(
            zip(
                [
                    "Frequency",
                    "FrequencyDerivative",
                    "Amplitude",
                    "EclipticLatitude",
                    "EclipticLongitude",
                    "Polarization",
                    "Inclination",
                    "InitialPhase",
                ],
                params,
            )
        )

    def to_fastgb_array(self) -> np.ndarray:
        """Create fastgb array, respecting format and ordering.

        The parameter array assumes the epoch `t_init`, and fastgb is doing the
        conversion between `tinit` and `t0` internally.

        Additional comments:

        fastgb expects that `params[:, 3]` contain ecliptic latitudes (beta) and
        `params[:, 4]` contain ecliptic longitudes (lambda). However, from LISA
        Orbits v3 onwards, orbit information is expressed in equatorial coordinates,
        so implictly, the sky location params in fastgb need to be expressed in
        equatorial coordinates. In principle, we can therefore pass the equatorial
        coordinates, as in JAXGB, provided we swap the longitude and latitude order:

        - In JAXGB, `params[:, 3]` is longitude (ra) while `params[:, 4]` includes
          latitude (dec),
        - In fastgb, `params[:, 3]` (what used to be beta) includes latitude, while
          `params[:, 4]` includes longitude (what used to be beta)

        Returns
        -------
        np.ndarray of shape (nsrc, 8)
            Array of parameters.
        """
        # fastgb applies internal conversion between `t0` and `t_init`, so we assume
        # them to be identical here
        params = self._to_array_at_reference_time(t0=self.t_init)
        return params[[0, 1, 2, 4, 3, 5, 6, 7]]

    def to_gwresponse_dict(self) -> dict[str, np.ndarray]:
        """Create GW Response input dictionary, respecting format and ordering.

        This can be used with :class:`lisagwresponse.GalacticBinary`. The parameter
        dictionary assumes the epoch `t_init`.

        Returns
        -------
            Dictionary of parameters, compatible with
            :class:`lisagwresponse.GalacticBinary`.
        """
        # GW response applies own internal time shift for `t0` and `t_init`, so we
        # assume them to be identical here
        params = self._to_array_at_reference_time(t0=self.t_init)
        return dict(
            zip(
                [
                    "f",
                    "df",
                    "A",
                    "ra",
                    "dec",
                    "psi",
                    "iota",
                    "phi0",
                ],
                params,
            )
        )

    def to_gbgpu_array(self) -> np.ndarray:
        """Create gbgpu array, respecting format and ordering.

        Sky positions are converted to ecliptic coordinates. ``fddot`` is
        assumed to be zero. The returned parameters have the epoch `t_init`.

        Returns
        -------
        np.ndarray of shape (nsrc, 9)
            Array of parameters.
        """
        # gbgbpu assumes no difference between `t0` and `t_init`
        params = self._to_array_at_reference_time(t0=self.t_init)
        params = params[[2, 0, 1, 1, 7, 6, 5, 3, 4]]
        params[3] = 0  # f2dot
        # convert equatorial to corresponding ecliptic coordinates
        params[7], params[8] = GBObject._equatorial_to_ecliptic(params[7], params[8])
        return np.array([np.full(1, p) for p in params])



"""Utilitary functions for testing fast waveform accuracy."""

import numpy as np
from jax import Array

from jaxgb.params import GBObject


def overlap(A1: Array, A2: Array) -> float:
    r"""Return overlap of two signals.

    .. math::

        O(A_1|A_2) = \frac{1}{\sqrt{ N_1 N_2}} \Re \int_{f_{\min}}^{f_{\max}}
        \Sum_i \tilde{A_1^i}(f) \tilde{A_2^i}^+(f) \mathrm{d}f,

    where

    .. :math::

        N^i = \int_{f_{\min}}^{f_{\max}} \tilde{A^i}^2(f) \mathrm{d}f}

    Parameters
    ----------
    A1
        TDI channels of 1st signal (nsamples, nchannels)
    A2
        TDI channels of 2nd signal (nsamples, nchannels)

    Returns
    -------
    float
        overlap of the 2 signals (between 0 and 1)
    """
    n1 = np.sum(np.sum(np.abs(A1) ** 2, axis=0), axis=0)
    n2 = np.sum(np.sum(np.abs(A2) ** 2, axis=0), axis=0)
    olap = np.sum(np.sum(np.real(A1 * A2.conj()), axis=0), axis=0)
    return np.abs(olap) / np.sqrt(n1 * n2)


def mismatch(A1: Array, A2: Array) -> float:
    r"""Return a well behaved mismatch (from the numerical point of view)

    .. math::

        m = \frac{1}{2} \left\Vert\frac{A_2}{\Vert A_2\Vert} -
        \frac{A_1}{\Vert A_1\Vert}\right\Vert^2

    Parameters
    ----------
    A1
        TDI channels of 1st signal (nsamples, nchannels)
    A2
        TDI channels of 2nd signal (nsamples, nchannels)

    Returns
    -------
    float
        mismatch of the 2 signals
    """
    n2 = A2 / np.sum(A2 * A2) ** 0.5
    n1 = A1 / np.sum(A1 * A1) ** 0.5
    return 0.5 * np.sum((n2 - n1) ** 2)


def get_source(
    name: str,
    as_dict: bool = False,
    as_gwresponse_dict: bool = False,
    as_gbgpu: bool = False,
    as_fastgb: bool = False,
) -> Array | dict:
    """Return source parameters from name.

    Parameters
    ----------
    name
        name of a known source (like `highfreq` for a high SNR, high frequency source).
    as_dict
        return parameters as a dictionary (LDC names)
    as_gwresponse_dict
        return parameters as a dictionary (lisagwresponse names)
    as_gbgpu
        return parameters as an array in the order expected by GBGPU
    as_fastgb
        return parameters as an array in the order expected by fastgb

    Returns
    -------
    array or dict of parameters of the chosen source.
    """
    gbobject = None
    if name == "highfreq":
        gbobject = GBObject(
            f0=0.018038,  # f0 Hz
            fdot=1.211329e-13,  # fdot "Hz^2
            A=2.765173e-22,  # ampl strain
            ra=4.22546,  # equatorial longitude (ra) radian
            dec=-0.823607,  # equatorial latitude (dec) radian
            psi=2.426183,  # polarization radian
            iota=1.052777,  # inclination radian
            phi0=3.020977,  # initial phase radian
            t_init=0.0,  # epoch in seconds
        )  # snr 130
    if as_dict:
        return gbobject.to_ldc_dict()
    if as_gwresponse_dict:
        return gbobject.to_gwresponse_dict()
    if as_gbgpu:
        return gbobject.to_gbgpu_array()
    if as_fastgb:
        return gbobject.to_fastgb_array()
    return gbobject.to_jaxgb_array(t0=0.0)



"""Compare JAXGB and GBGPU fast waveforms."""

import jax
import jax.numpy as jnp
import lisaorbits
import numpy as np
from fastgb import fastgb
from gbgpu.gbgpu import GBGPU
from testutils import get_source, overlap

from jaxgb.jaxgb import JaxGB
from jaxgb.tdi import to_tdi_combination

# pylint: disable=too-many-locals
jax.config.update("jax_enable_x64", True)


def test_gbgpu(gw_name: str, tdi_generation: float) -> None:
    """Check that overlap between JaxGB and GBGPU

    We assume equal armlength orbits.
    """
    size, dt, tmax = 2048, 15, 180 * 24 * 3600

    # GBGPU
    gb = GBGPU(force_backend="cpu")
    params = get_source(gw_name, as_gbgpu=True)
    gb.run_wave(*params, N=size, dt=dt, T=tmax, oversample=1, tdi2=tdi_generation == 2)

    A1 = gb.A[0]
    E1 = gb.E[0]
    freqs = gb.freqs[0]

    fgb = JaxGB(orbits=lisaorbits.EqualArmlengthOrbits(), t_obs=tmax, t0=0, n=size)
    params = get_source(gw_name)
    A2, E2, _ = fgb.get_tdi(
        params,
        tdi_generation=tdi_generation,
        tdi_combination="AET",
    )
    kmin = fgb.get_kmin(params[0])  # frequency

    assert A1.shape == A2.shape
    df = 1 / tmax
    assert jnp.isclose(freqs, df * jnp.arange(kmin, kmin + A2.shape[0])).all()
    olap = overlap(np.hstack([A1, E1]), np.hstack([A2, E2]))
    assert olap > 0.965


def test_fastgb(gw_name: str, tdi_generation: float) -> None:
    """Check that overlap between JaxGB and fastGB

    We assume equal armlength orbits.
    """
    size, dt, tmax = 2048, 15, 180 * 24 * 3600

    # fastGB
    fgb = fastgb.FastGB(
        delta_t=dt, T=tmax, N=size, orbits=lisaorbits.EqualArmlengthOrbits()
    )
    params = get_source(gw_name, as_fastgb=True)
    X, Y, Z, _ = fgb.get_fd_tdixyz(
        jnp.array(params).reshape(1, -1), tdi2=tdi_generation == 2
    )
    tdi = to_tdi_combination(jnp.stack([X, Y, Z], axis=1), "AET")
    A1, E1 = tdi[:, 0, :], tdi[:, 1, :]

    # JAXGB
    params = get_source(gw_name)
    jgb = JaxGB(orbits=lisaorbits.EqualArmlengthOrbits(), t_obs=tmax, t0=0, n=size)
    A2, E2, _ = jgb.get_tdi(
        params,
        tdi_generation=tdi_generation,
        tdi_combination="AET",
    )

    olap = overlap(np.hstack([A1, E1]), np.hstack([A2, E2]))
    assert olap > 0.9999



"""
Example module
==============

This is some description.

Some functions
--------------

.. autofunction:: hello_world

"""


def hello_world(name: str) -> None:
    """Greets someone.

    Parameters
    ----------
    name
        Person to greet.
    """
    print(f"Hello, {name}!")



"""Compare fast waveform to time domain generated waveform using
gwresponse and pytdi.

The comparison is performed in the frequency domain.
"""

import os

import h5py
import jax
import jax.numpy as jnp
import lisagwresponse
import lisaorbits
import numpy as np
import pandas as pd
import pytdi.michelson
import pytest
from jax import Array
from lisaconstants.indexing import LINKS
from lisaorbits import ResampledOrbits
from pytdi import Data
from testutils import get_source, mismatch, overlap

from jaxgb.jaxgb import JaxGB, JaxGBaccurate
from jaxgb.params import GBObject
from jaxgb.tdi import to_tdi_combination

jax.config.update("jax_enable_x64", True)

# pylint: disable=too-many-locals
# pylint: disable=redefined-outer-name


def _read_t0(orbits_file):
    with h5py.File(orbits_file, "r") as fid:
        t0 = fid.attrs["t0"]
    return t0


def _gw2tdi(
    orbits_fn: str,
    gwfile: str,
    dt: int,
    tdi_generation: float = 2,
    tdi_combination: str = "XYZ",
) -> tuple[Array, Array, Array, Array]:
    """Return TDI X,Y,Z channel and associated frequencies from link
    response stored in `gwfile`.
    """
    data = Data.from_gws(gwfile, orbits_fn)
    if tdi_generation == 2:
        X = pytdi.michelson.X2.build(**data.args)(data.measurements)
        Y = pytdi.michelson.Y2.build(**data.args)(data.measurements)
        Z = pytdi.michelson.Z2.build(**data.args)(data.measurements)
    else:
        X = pytdi.michelson.X1.build(**data.args)(data.measurements)
        Y = pytdi.michelson.Y1.build(**data.args)(data.measurements)
        Z = pytdi.michelson.Z1.build(**data.args)(data.measurements)

    ## Move to freq domain
    freqs = np.fft.rfftfreq(X.size, d=dt)
    X = np.fft.rfft(X) * dt
    Y = np.fft.rfft(Y) * dt
    Z = np.fft.rfft(Z) * dt
    tdi = to_tdi_combination(
        jnp.stack([X.reshape(1, -1), Y.reshape(1, -1), Z.reshape(1, -1)], axis=1),
        tdi_combination,
    )
    return freqs, tdi[:, 0, :].squeeze(), tdi[:, 1, :].squeeze(), tdi[:, 2, :].squeeze()


def test_tdi_overlap(
    tmp_path: str | os.PathLike,
    gw_name: str,
    tdi_generation: str,
    orbits_approximation: str,
) -> None:
    """Check that overlap between fast waveform and time domain waveform
    is above 0.999 for a given source, orbits approximation and tdi
    generation.

    """
    size, dt, tmax = 2048, 15, 180 * 24 * 3600

    ## compute GW response and TDI
    gwfile = str(tmp_path / "gw.h5")
    p = get_source(gw_name, as_gwresponse_dict=True)
    p["orbits"] = orbits_approximation
    gb = lisagwresponse.GalacticBinary(**p)
    gb.write(gwfile, mode="w", t0=0, size=tmax // dt, dt=dt)
    freqs, A1, E1, _ = _gw2tdi(
        orbits_approximation,
        gwfile,
        dt,
        tdi_generation=tdi_generation,
        tdi_combination="AET",
    )

    ## compute fast approx
    fgb = JaxGB(orbits=ResampledOrbits(orbits_approximation), t_obs=tmax, t0=0, n=size)
    params = get_source(gw_name)
    A2, E2, _ = fgb.get_tdi(
        params, tdi_generation=tdi_generation, tdi_combination="AET"
    )
    kmin = fgb.get_kmin(params[0])  # frequency

    ## align on A/E
    df = 1 / tmax
    i0 = np.where(freqs == df * kmin)[0][0]
    sl = slice(i0, i0 + A2.shape[0])
    olap = overlap(np.hstack([A1[sl], E1[sl]]), np.hstack([A2, E2]))
    assert olap > 0.999


def test_link_overlap(gw_name: str, orbits_approximation: str) -> None:
    """Check that overlap at the link level, between fast waveform and
    time domain waveform is above 0.999 for a given source, orbits
    approximation and tdi generation.
    Each link is tested independently, to avoid mismatch cancellation.

    """
    size, dt, tmax = 2048, 15, 180 * 24 * 3600

    ## compute GW response
    p = get_source(gw_name, as_gwresponse_dict=True)
    p["orbits"] = orbits_approximation
    gb = lisagwresponse.GalacticBinary(**p)
    times = np.arange(0, tmax, dt).astype(float)
    yt = gb.compute_gw_response(times, LINKS)
    yf1 = np.fft.rfft(yt.T) * dt
    freqs = np.fft.rfftfreq(yt.shape[0], d=dt)

    ## compute fast response
    fgb = JaxGB(orbits=ResampledOrbits(orbits_approximation), t_obs=tmax, t0=0, n=size)

    params = get_source(gw_name)
    yf2 = fgb.get_link_responses(params.reshape(1, -1))[0, :, :]
    kmin = fgb.get_kmin(params[0])  # frequency

    ## align
    df = 1 / tmax
    i0 = np.where(freqs == df * kmin)[0][0]
    sl = slice(i0, i0 + yf2.shape[1])
    for link in range(6):
        olap = overlap(yf1[link, sl], np.array(yf2[link, :]))
        assert olap > 0.999


def test_esa_orbits(
    tmp_path: str | os.PathLike,
    gw_name: str,
    tdi_generation: str,
) -> None:
    """Check that overlap between fast waveform and time domain waveform
    is above 0.99 when using esa orbits. Handling t0!=0 case.
    """
    # esa orbits
    tmax = 730 * 24 * 3600
    orbits = lisaorbits.OEMOrbits.from_included("esa-leading")
    t0 = orbits.t_init + 50000000
    orbits_fn = str(tmp_path / "orbits.h5")
    orbits.write(orbits_fn, mode="w", dt=1e5, size=(2 * tmax) // 1e5, t0=t0 - 100)

    size, dt, tmax = 4096, 15, 180 * 24 * 3600

    ## compute GW response and TDI
    gwfile = str(tmp_path / "gw.h5")
    p = get_source(gw_name, as_gwresponse_dict=True)
    p["orbits"] = orbits_fn
    p["t_init"] = t0
    gb = lisagwresponse.GalacticBinary(**p)
    gb.write(gwfile, mode="w", t0=t0, size=tmax // dt, dt=dt)

    freqs, A1, E1, _ = _gw2tdi(
        orbits_fn, gwfile, dt, tdi_generation=tdi_generation, tdi_combination="AET"
    )

    ## compute fast approx
    fgb = JaxGB(orbits=orbits, t_obs=tmax, t0=t0, n=size)
    params = get_source(gw_name)
    A2, E2, _ = fgb.get_tdi(
        params, tdi_generation=tdi_generation, tdi_combination="AET"
    )
    kmin = fgb.get_kmin(params[0])  # frequency

    ## align on A/E
    df = 1 / tmax
    i0 = np.where(freqs == df * kmin)[0][0]
    sl = slice(i0, i0 + A2.shape[0])
    olap = overlap(np.hstack([A1[sl], E1[sl]]), np.hstack([A2, E2]))
    assert olap > 0.99


def test_accurate(gw_name: str, orbits_approximation: str) -> None:
    """Check overlap of accurate fast waveform versus time domain at the
    link level, in time domain.
    """
    size, dt, tmax = 4096, 15, 180 * 24 * 3600
    t0 = 0

    ## fast response
    fgb = JaxGBaccurate(
        orbits=ResampledOrbits(orbits_approximation),
        t_obs=tmax,
        t0=t0,
        n=size,
        window=0.5,
    )
    params = get_source(gw_name)
    yf = fgb.get_link_responses(params.reshape(1, -1))[0, :, :]
    kmin = fgb.get_kmin(params[0])  # frequency

    ## time domain
    p = get_source(gw_name, as_gwresponse_dict=True)
    p["orbits"] = orbits_approximation
    p["t_init"] = t0
    gb = lisagwresponse.GalacticBinary(**p)
    times = np.arange(t0, t0 + fgb.t_obs, dt).astype(float)
    yt = gb.compute_gw_response(times, LINKS)

    df = 1 / fgb.t_ext
    target_length = int(np.rint(1 / df / dt))

    for link in range(6):
        array = np.pad(yf[link, :], (kmin, 0))
        yt_fast = np.fft.irfft(array / dt, n=target_length)
        times_fast = fgb.tm[0] + np.arange(target_length) * dt

        # align
        i0 = np.where(times_fast == t0)[0][0]
        i1 = i0 + len(times)
        olap = overlap(yt[:, link], yt_fast[i0:i1])
        assert olap > 0.999
        mm = mismatch(yt[:, link], yt_fast[i0:i1])
        assert mm < 5e-15


@pytest.mark.parametrize("t_init", [97729930.202664, 50729089.327664])
def test_tinit(tmp_path: str | os.PathLike, t_init: float) -> None:
    """test t_init = t0 and t_init != t0

    a huge time difference is needed to be sensitive to the phase
    shift.
    """

    t0 = 97729930.202664

    src = pd.DataFrame(
        {
            "Amplitude": [4.184033623373269e-23],
            "RightAscension": [4.577818480741154],
            "Declination": [-0.4451836132523424],
            "Polarization": [2.1401202483554256],
            "Inclination": [2.5150470519327444],
            "FrequencyDerivative": [3.425955521564606e-16],
            "Frequency": [0.006094623731395194],
            "InitialPhase": [0.8363015305549726],
        }
    )
    gbo = GBObject.from_pandas_dataframe(src, t_init=t_init)
    params0 = gbo.to_jaxgb_array(t0=t_init)  # to be given to gw_response, no shift
    params = gbo.to_jaxgb_array(t0=t0)  # to be given to jaxgb, with shift

    duration = 24 * 3600 * 30 * 4
    orbits_sampling = {
        "dt": np.float64(500000.0),
        "duration": np.float64(137000000.0),
        "fs": np.float64(2e-06),
        "size": np.int64(274),
        "t0": np.float64(61171239.327664),
    }

    # esa orbits
    orbits_fn = str(tmp_path / "orbits.h5")
    # _orbits = lisaorbits.OEMOrbits.from_included("esa-trailing")
    _orbits = lisaorbits.EqualArmlengthOrbits()

    _orbits.write(
        orbits_fn,
        mode="w",
        dt=orbits_sampling["dt"],
        size=orbits_sampling["size"],
        t0=orbits_sampling["t0"],
    )

    size, dt = 1024, 15

    ## compute GW response and TDI
    gwfile = str(tmp_path / "gw.h5")
    p = dict(
        zip(
            [
                "f",
                "df",
                "A",
                "ra",
                "dec",
                "psi",
                "iota",
                "phi0",
            ],
            [float(p) for p in params0.squeeze()],
        )
    )
    p["orbits"] = orbits_fn
    p["t_init"] = t_init
    gb = lisagwresponse.GalacticBinary(**p)
    gb.write(gwfile, mode="w", t0=t0, size=duration // dt, dt=dt)

    freqs, A1, E1, _ = _gw2tdi(
        orbits_fn, gwfile, dt, tdi_generation=2, tdi_combination="AET"
    )

    ## compute fast approx
    fgb = JaxGB(orbits=_orbits, t_obs=duration, t0=t0, n=size)

    A2, E2, _ = fgb.get_tdi(params, tdi_generation=2, tdi_combination="AET")
    kmin = fgb.get_kmin(params[0, 0])  # frequency

    ## align on A/E
    df = 1 / fgb.t_obs
    i0 = np.where(freqs == df * kmin)[0][0]
    sl = slice(i0, i0 + A2.shape[1])
    olap = overlap(np.hstack([A1[sl], E1[sl]]), np.hstack([A2, E2]))
    assert olap > 0.99
